<?php

namespace App\Command;

use LogicException;
use phpDocumentor\Reflection\Types\Array_;
use phpDocumentor\Reflection\Types\Boolean;
use Symfony\Component\Console\Attribute\AsCommand;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Helper\ProgressBar;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\ConsoleOutputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Filesystem\Path;
use Symfony\Component\HttpKernel\KernelInterface;

#[AsCommand(
    name: 'rekall:projects:update',
    description: 'Update commons files on projects.',
    aliases: ['rekall-update-projects'],
    hidden: false
)]
class UpdateAllProjects extends Command
{
    private Filesystem $file_system;
    private KernelInterface $kernel;

    private string $legacy_root;
    private string $capsule_prototype_root;

    /**
     * @var string[] $BLACK_LIST files and directories that should not be taken into account for projects update
     */
    private static array $BLACK_LIST =
        ['css', 'create.zip', 'favicon.ico', 'la-page-des-malins.php', 'js', 'php',  'shared', '.', '..'];
    private static string $CAPSULE_PROTOTYPE_DIRECTORY_NAME = 'capsule-prototype';
    private static string $LEGACY_DIRECTORY_NAME = 'legacy';


    public function __construct(Filesystem $filesystem, KernelInterface $kernel)
    {
        parent::__construct();

        // best practices recommend to call the parent constructor first and
        // then set your own properties. That wouldn't work in this case
        // because configure() needs the properties set in this constructor
        $this->file_system = $filesystem;
        $this->kernel = $kernel;
        $this->legacy_root = Path::join($this->kernel->getProjectDir(), self::$LEGACY_DIRECTORY_NAME);
        $this->capsule_prototype_root = Path::join(
            $this->kernel->getProjectDir(),
            self::$CAPSULE_PROTOTYPE_DIRECTORY_NAME
        );
    }

    protected function configure(): void
    {
        $this
            // If you don't like using the $defaultDescription static property,
            // you can also define the short description using this method:
            ->setDescription('Update every project php files and index.html')

            // the command help shown when running the command with the "--help" option
            ->setHelp('This command allows update projects common files...');
    }

    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        if (!$output instanceof ConsoleOutputInterface) {
            throw new LogicException('This command accepts only an instance of "ConsoleOutputInterface".');
        }

        ProgressBar::setFormatDefinition(
            'project_update_message',
            ' %current%/%max% [%bar%] %percent%% : Updating project :%project% %elapsed%'
        );
        $output->writeln('Updating all projects started');

        $all_file_in_directory = $this->getProjectsDirectories();
        if (false === $all_file_in_directory) {
            return Command::FAILURE;
        }

        $php_files_count_per_projects = $this->getPhpFilesCountPerProjects();

        $all_projects_progress_bar = new ProgressBar($output->section());
        $all_projects_progress_bar->setFormat(
            '%current%/%max% [%bar%] %percent%% : Overall project update progressing (%elapsed%)'
        );
        $all_projects_progress_bar->SetMessage('Projects update progress');
        $all_projects_progress_bar->start();

        foreach ($all_projects_progress_bar->iterate($all_file_in_directory) as $project_directory) {
            $current_legacy_path = Path::join($this->legacy_root, $project_directory);

            if (!$this->file_system->exists($current_legacy_path)) {
                continue;
            }

            if (!is_dir($current_legacy_path)) {
                continue;
            }

            if (in_array($project_directory, self::$BLACK_LIST)) {
                continue;
            }

            $this->updateProjectDirectory(
                $output,
                $current_legacy_path,
                $this->capsule_prototype_root
            );
            $all_projects_progress_bar->advance($php_files_count_per_projects);
        }

        $all_projects_progress_bar->finish();

        return Command::SUCCESS;
    }

    /**
     * @return string[]|false
     */
    private function getProjectsDirectories(): array|false
    {

        $all_directories_and_files = scandir($this->legacy_root);
        if (false === $all_directories_and_files) {
            return false;
        }
        return array_diff($all_directories_and_files, self::$BLACK_LIST);
    }

    /**
     * @return int
     */
    protected function getPhpFilesCountPerProjects(): int
    {
        $php_files = glob($this->capsule_prototype_root . '/php/*.php');
        // add +1 for index.html file
        return count($php_files ?: []) + 1;
    }

    /**
     * @param ConsoleOutputInterface $output
     * @param string $current_legacy_path
     * @param string $capsule_prototype_root
     * @return void
     */
    protected function updateProjectDirectory(
        ConsoleOutputInterface $output,
        string $current_legacy_path,
        string $capsule_prototype_root
    ): void {
        $output->writeln("Updating project :{$current_legacy_path}");

        // override index.html
        $this->file_system->copy(
            Path::normalize($capsule_prototype_root . '/index.html'),
            Path::normalize($current_legacy_path . '/index.html'),
            true
        );

        // override php files
        $this->file_system->mirror(
            Path::join($capsule_prototype_root, 'php'),
            Path::join($current_legacy_path, 'php'),
            null,
            ['override' => true, 'delete' => true]
        );

        $output->writeln("Updating project :{$current_legacy_path} ---> DONE");
    }
}
