<?php

namespace App\Controller;

use App\Entity\PendingEmailAddress;
use App\Entity\User;
use App\Form\ChooseLanguageFormType;
use App\Form\EditPasswordFormType;
use App\Form\EditUserProfileFormType;
use App\Repository\PendingEmailAddressRepository;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Bridge\Twig\Mime\TemplatedEmail;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Mailer\MailerInterface;
use Symfony\Component\Mime\Address;
use Symfony\Component\PasswordHasher\Hasher\UserPasswordHasherInterface;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Contracts\Translation\TranslatorInterface;

class UserController extends AbstractController
{
    public function __construct(
        private EntityManagerInterface $entity_manager,
        private TranslatorInterface $translator,
        private PendingEmailAddressRepository $pending_email_address_repository
    ) {
    }

    #[Route('/{_locale<%app.supported_locales%>}/profile', name:'show_profile')]
    public function showProfile(): Response
    {
        $current_user = $this->getUser();

        if (! $current_user instanceof User) {
            return $this->redirectToRoute('app_logout');
        }

        return $this->render('user/profile.html.twig', [
            'user' => $current_user
        ]);
    }

    #[Route('/{_locale<%app.supported_locales%>}/edit_profile', name:'edit_profile')]
    public function editProfile(Request $request, MailerInterface $mailer): Response
    {
        $current_user = $this->getUser();

        if (! $current_user instanceof User) {
            return $this->redirectToRoute('app_logout');
        }

        $last_pending_email_address = $this->pending_email_address_repository->findOneBy(
            ['user' => $current_user->getId()]
        );
        if ($last_pending_email_address !== null) {
            $this->entity_manager->remove($last_pending_email_address);
        }

        $form = $this->createForm(
            EditUserProfileFormType::class,
            $current_user,
            ['current_email_address' => $current_user->getEmail()]
        );
        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid()) {
            if ($current_user->getEmail() !== $form->get('email')->getData()) {
                $pending_email_address = new PendingEmailAddress();
                $pending_email_address->setEmail($form->get('email')->getData());
                $pending_email_address->setUser($current_user);
                $this->entity_manager->persist($pending_email_address);
                $this->entity_manager->flush();

                $email = (new TemplatedEmail())
                    ->to(new Address($pending_email_address->getEmail()))
                    ->subject($this->translator->trans('user.edit.email.title'))
                    ->htmlTemplate('user/update_email.html.twig')
                    ->context([
                        'expiration_date' => new \DateTime('+1 day')
                    ]);
                $mailer->send($email);

                $this->addFlash(
                    'warning',
                    $this->translator->trans('user.profile.updated.warning', [
                        'new_email_address' => $form->get('email')->getData()
                    ])
                );

                return $this->render('user/edit_email_address.html.twig', [
                    'new_email_address' => $form->get('email')->getData()
                ]);
            }

            $this->entity_manager->persist($current_user);
            $this->entity_manager->flush();

            $this->addFlash(
                'success',
                $this->translator->trans('user.profile.updated.success')
            );

            return $this->redirectToRoute('show_profile');
        }

        return $this->renderForm('user/edit_profile.html.twig', [
            'editUserProfileForm' => $form
        ]);
    }

    #[Route('/{_locale<%app.supported_locales%>}/edit_password', name:'edit_password')]
    public function editPassword(
        Request $request,
        UserPasswordHasherInterface $user_password_hasher
    ): Response {
        $form = $this->createForm(EditPasswordFormType::class);
        $form->handleRequest($request);
        $current_user = $this->getUser();

        if (! $current_user instanceof User) {
            return $this->redirectToRoute('app_logout');
        }
        if ($form->isSubmitted() && $form->isValid()) {
            $encodedPassword = $user_password_hasher->hashPassword(
                $current_user,
                $form->get('plainPassword')->getData()
            );

            $current_user->setPassword($encodedPassword);
            $this->entity_manager->flush();

            $this->addFlash(
                'success',
                $this->translator->trans('user.password.updated_success')
            );

            return $this->redirectToRoute('show_profile');
        }
        return $this->renderForm('user/edit_password.html.twig', [
            'editPasswordForm' => $form
        ]);
    }

    #[Route('/{_locale<%app.supported_locales%>}/edit_language', name:'choose_language')]
    public function editLanguage(Request $request): Response
    {
        $current_user = $this->getUser();

        if (! $current_user instanceof User) {
            return $this->redirectToRoute('app_logout');
        }

        $form = $this->createForm(ChooseLanguageFormType::class, null, ['locale' => $current_user->getLocale()]);
        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid()) {
            $locale = $form->get('language')->getData();
            $request->setLocale($locale);
            $current_user->setLocale($locale);
            $this->entity_manager->persist($current_user);
            $this->entity_manager->flush();

            $this->addFlash(
                'success',
                $this->translator->trans(
                    'user.profile.language.updated_success',
                    [],
                    null,
                    $locale
                )
            );

            return $this->redirectToRoute('show_profile', ['_locale' => $locale]);
        }
        return $this->renderForm('user/edit_language.html.twig', [
            'chooseLanguageForm' => $form
        ]);
    }

    #[Route('/edit/email', name:'verify_new_email_address')]
    public function verifyNewEmailAddress(): Response
    {
        $current_user = $this->getUser();
        if (! $current_user instanceof User) {
            return $this->redirectToRoute('app_logout');
        }

        $pending_email_address = $this->pending_email_address_repository->findOneBy(['user' => $current_user->getId()]);
        if (! $pending_email_address instanceof PendingEmailAddress) {
            throw new \Exception('Pending email address not found');
        }

        $current_user->setEmail($pending_email_address->getEmail());
        $this->entity_manager->persist($current_user);
        $this->entity_manager->remove($pending_email_address);
        $this->entity_manager->flush();

        $this->addFlash(
            'success',
            $this->translator->trans(
                'user.edit.email.success',
                ['new_email_address' => $pending_email_address->getEmail()]
            )
        );

        return $this->redirectToRoute('show_profile');
    }
}
